#------------------------------------------------------------------------------------------------
# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#------------------------------------------------------------------------------------------------
#   Name:	EdgeFollowMe
#     By:	sdmitch
#  Usage:	Simulates Follow Me using an edge instead of a face.
#	  Note:	The profile to be extruded, must be a curve.  Use Weld or simuliar plugin to combine the
#					entities defining the profile.  The profile should be at the start of the extrusion path.
#					If a path is pre-selected, the start and end will be labled.  Use right or left arrow keys
#					to reverse the line.  If a path is not pre-selected, click on the profile and drag to 
#					extrude the profile to any length and direction.
#    			Date:	Aug 2011
#------------------------------------------------------------------------------------------------

require 'Sketchup'

filename=File.basename(__FILE__)
if !file_loaded?(filename)
	UI.menu("Plugins").add_item("Edge Follow Me") { Sketchup.active_model.select_tool Edge_Follow_Me.new }
end
file_loaded(filename)

# ============================================================================

class Edge_Follow_Me

	def initialize
		@ip0 = Sketchup::InputPoint.new
		@ip1 = Sketchup::InputPoint.new
		@mod= Sketchup.active_model
		@ent = @mod.active_entities
		@sel = @mod.selection
		self.reset
		if !@sel.empty?
			if @sel.first.curve
				@path=@sel.first.curve.edges
				verts=@sel.first.curve.vertices
				@pts=verts.collect{|v| v.position}
			else
				@path,@pts=sort_edges(@sel,0)
			end
			@mod.start_operation "Edge Follow Me"
			self.lable_line
			@sel.clear; 
		end
		view = @mod.active_view.lock_inference
	end

	def reset
		@state = 0
		@path = []
		@best = nil
		@edge = nil
		@ip1.clear
		@beg_pts=[]
		@end_pts=[]
		@drawn = false
	end

	def onMouseMove(flags, x, y, view)
		self.set_current_point(x, y, view)
		ph = view.pick_helper
		ph.do_pick(x, y)
		@edge = ph.picked_edge if ph.picked_edge != @edge
		@best = ph.best_picked if ph.best_picked != @best
		Sketchup::set_status_text("Select Profile Curve", SB_PROMPT) if @state == 0
		Sketchup::set_status_text("Drag to Extrude Edge", SB_PROMPT) if @state == 1
	end

	def set_current_point(x, y, view)
		if( !@ip0.pick(view, x, y, @ip1))
			return false
		end
		@ip1.copy! @ip0
		view.tooltip = @ip0.tooltip     
		view.refresh
	end

	def onLButtonDown(flags, x, y, view)
		if @state == 0 then
			if (@edge && @best.is_a?(Sketchup::Edge)) then
				if @best.curve
					@sel.add @best.curve.edges
					verts = @best.curve.vertices
					verts.each{|v| @end_pts<<v.position}
					@line=@best.line;@state = 1
					if @path.length > 0
						if @end_pts.include?(@pts[0])
							self.extrude_edge 
						else
							if UI.messagebox("Does not appear to be a valid path.  Continue?",MB_YESNO)==6
								self.extrude_edge 
							end
						end
						@sel.clear; @ent.erase_entities(@start_label,@end_label)
						Sketchup.send_action "selectSelectionTool:"
						@mod.commit_operation
					end
				else
					UI.messagebox "The profile must be a curve"
				end
			end
		elsif @state == 1
			pt0 = @ip0.position; pt1=pt0.project_to_line @line
			vec=pt1.vector_to pt0; dis=pt1.distance pt0
			@mod.start_operation "Edge Follow Me"
			@pts=[];@pts[0]=@end_pts[0];@pts[1]=@end_pts[0].offset(vec,dis)
			@path[0]=@ent.add_line(@pts[0],@pts[1])
			self.extrude_edge
			@sel.clear;self.reset
		end
	end
	
	def extrude_edge
		last = @path.length - 1
		@grp=@ent.add_group; @gents=@grp.entities
		sas=UI.messagebox("Soften and smooth the surface?",MB_YESNO)==6
		for n in 0..last
			pt1=@pts[n];vec1=@pts[n].vector_to(@pts[n+1])
			if n < last
				pt2=@pts[n+1];vec2=@pts[n+1].vector_to(@pts[n+2])
				vec=Geom::Vector3d.linear_combination(0.5,vec1,0.5,vec2)
			else
				pt2=@pts[-1]
				vec=vec1
			end
			plane=[pt2,vec]
			for i in 0...@end_pts.length
				@beg_pts[i]=@end_pts[i]; line = [@beg_pts[i],vec1]
				@end_pts[i]=Geom.intersect_line_plane(line,plane)
			end
			for i in 1...@end_pts.length
				face=@gents.add_face(@beg_pts[i-1],@end_pts[i-1],@end_pts[i],@beg_pts[i])
				face.edges.each{|e| (e.soft=true;e.smooth=true) if e.faces.length>1} if sas
			end
		end
	end
	
	def onCancel(flag, view)
		view.refresh if @drawn
		Sketchup.send_action "selectSelectionTool:"
	end
	
	def onRButtonDown(flags, x, y, view)
		@mod.select_tool(nil) if @state==0
		self.reset
	end
	
	def onKeyDown(key, repeat, flags, view)
		if (key == VK_RIGHT || key == VK_LEFT)
			@ent.erase_entities(@start_label,@end_label)
			@pts.reverse!
			self.lable_line
		end
	end
	
	def lable_line
		@start_label=@ent.add_text("Start",@pts[0])
		@end_label=@ent.add_text("End",@pts[-1])
	end
	
	def draw(view)
		@drawn = false
		if( @ip0.valid? && @ip0.display? ) then
			@ip0.draw(view)
			@drawn = true
		end
		if @state == 0 && @edge
			if @edge.curve
				view.line_width=3
				view.drawing_color = "blue"
				verts=@edge.curve.vertices; pts=[]
				verts.each{|v| pts<<v.position}
				view.draw(GL_LINE_STRIP,pts)
			end
		elsif @state > 0 && @path.length == 0
			view.drawing_color = "magenta"
			view.line_width = 2
			pt0 = @ip0.position; pt1=pt0.project_to_line @line
			vec=pt1.vector_to pt0; dis=pt1.distance pt0
			if dis > 1.0
				for i in 1...@end_pts.length
					p0=@end_pts[i-1].offset(vec,dis); 
					p1=@end_pts[i].offset(vec,dis)
					view.draw(GL_LINE_LOOP,@end_pts[i-1],p0,p1,@end_pts[i])
				end
			end
		end
	end
	
	def sort_edges(sel,flag)
	#
	#	Sort edges array and returns the edges and vertices.
	#
		edges=sel.select {|x| x.is_a? Sketchup::Edge}
	#	---------------- This code copied from FAK.rb ------------------
		return if edges.length==0
		re=[];rp=[]	# split the output into two arrays
		while edges.length!=0
			start=nil
			edges.each{|x|
			start=x.start; break  if (x.start.edges&edges).length==1
			start=x.end; break if (x.end.edges&edges).length==1
			}
			vts=[]; curve=[]
			while start
				next_edge=(start.edges&edges-curve)[0]
				if next_edge
					curve << next_edge
					vts << start
					start=next_edge.other_vertex(start)
				else
					vts << start
					start=nil
				end
			end
			edges-=curve
			re += curve 
			rp += vts 
		end
	#	----------------------------------------------------------------
		rp.collect!{|x| x.position}
		return rp if flag==2
		return re if flag==1
		return re,rp	#re=edges, rp=points
	end
	
end # of class

# ============================================================================
